#include "urn_graph.hpp"
#include "algorithms.hpp"
#include "fastlog.h"


Urn::Urn(int N0, double p0, double gamma0, double Nu, double Rho, double Rho0){


	struct timeval tv;
	gettimeofday(&tv,NULL);
	init_genrand((tv.tv_sec * 1000) + (tv.tv_usec / 1000));
	//init_genrand(10045454);
	srand((tv.tv_sec * 1000) + (tv.tv_usec / 1000));

	engine = default_random_engine(random_device{}());

	/*init_genrand(0);
	srand(0);
	engine = default_random_engine(0);*/


	p = p0;
	nu = Nu;
	rho = Rho;
	gamma = gamma0;
	N_balls = N0/(nu+1);

	N_links = 0;
	for(int i = 0; i < N_balls; i++){
		balls.push_back(i);
		children_number.push_back(0);
		labels_count.push_back(0);
		weights.push_back(1);
		adj_list.push_back(vector<int>(0));
		set_random_links_full(i);
		N_links+=adj_list[i].size();
		exited.push_back(false);
		labels.push_back(-1);
	}
}

Urn::~Urn(){}

void Urn::add_ball(int source){
	balls.push_back(N_balls);
	weights.push_back(1);
	labels.push_back(source);

	adj_list.push_back(vector<int>(0));
	children_number.push_back(0);
	labels_count.push_back(0);
	/*for(int i =0 ; i < N_balls; i++)
		if(u() <= p or i == source) {
			adj_list[i].push_back(N_balls);N_links+=1;
			adj_list[N_balls].push_back(i);N_links+=1;
		}

	*/

	adj_list[N_balls].push_back(N_balls);
	adj_list[N_balls].push_back(source);
	labels_count[labels[N_balls]+1]+=1;
	adj_list[source].push_back(N_balls);

	//set_random_links(N_balls);
	N_links+=adj_list[N_balls].size();
	N_balls++;
	exited.push_back(false);

}

void Urn::set_random_links_full(int c){
	for(int i =0 ; i < N_balls; i++) adj_list[c].push_back(i);
}


void Urn::reinforce_ball(int c, double w){
	weights[c]+=w;
}

int Urn::get_starting_ball(){

	int q_start = u(N_balls);
	reinforce_ball(q_start, rho);
	children_number[q_start] = nu +1;
	creation_time.push_back(0);
	last_visit.push_back(0);
	//for(int n = 0; n < nu; n++) add_ball(q_start);
	exited[q_start] = true;
	labels_count[labels[q_start]+1]+=1;
	int extract_b = balls[q_start];
	return extract_b;
	
}

int Urn::get_ball(int c, int t){

	vector<int> next_list = adj_list[c];

	vector<bool> linked(N_balls, false);
	vector<double> children_connection(N_balls, 0);
	for(int i = 0; i < next_list.size();i++){
		int b = next_list[i];
		if(not linked[b]){
			linked[b] = true;
		}
	}


	double N_k = 0, N_nk = 0, max_N_k = 0, tot_k = 0;
	for(int i = 0; i < N_balls; i++){
		if(labels[i]==labels[c]) N_k+=weights[i];
		else N_nk+=weights[i];
	}

	

	/*double g = pow(N_k/(N_k + gamma*N_nk),0.95);
	if(g<0) g = 0;
	double f = (g*(N_k+gamma*N_nk) - N_k)/gamma/N_nk;
	if(g==0 or N_nk==0) f = 0;*/
	double f = fastPow(N_k/(N_k + N_nk),1);
	double g = (N_k + gamma*f*N_nk)/(N_k + gamma*N_nk);
	if(f>1) f = 1;
	if(f<0){
		cout << "errore!!"<< endl;
		cout << "g = "<< g << " f = "<< f << endl;
		cout << "g_lim = " << N_k/(N_k + gamma*N_nk) << endl;
		cout << "g_power = " << pow(N_k/(N_k + gamma*N_nk),0.9) << endl;

		exit(-1);
	}

	//f=1, g = 1;
	if(g>1) g = 1;
	if(f>1) f = 1;

	if((t-1)%10000==0) {
		cout <<N_k << "\t"<< N_nk << "\t" << f << "\t" << g << "\t" << 1.0/p/N_balls << endl;
		//exit(-1);
	}

	for(int i = 0; i < N_balls; i++){
		if(labels[i] == labels[c]) linked[i] = true;
		if(labels[i] == c) linked[i] = true;

	}


	double norm = 0;
	for(int i = 0; i < N_balls; i++){
		if(children_number[i]>nu+1){
			cout << "Mierda!!!\n";
			exit(-1);
		}
		int b = balls[i];
		double w = weights[i];
		if(linked[i]) norm+=w;			
		else if(labels[i]==labels[c]) norm+=w;
		else norm+=w*gamma*f;

		if(children_number[i]>0){
			if(i==c){
				double to_add = (children_number[i])*g;
				/*if(to_add<1 and children_number[i]>0) 
					to_add = 1.0;*/
				norm+=to_add;			
			}
			else if(i==labels[c]){
				/*double to_add = children_number[i]*g;
				if(to_add<1 and children_number[i]>0) 
					to_add = 1.0;*/

				double to_add = (children_number[i])*g;
				norm+=to_add;
			}
			else norm+=children_number[i]*g*p;
		}
		//cout << i << "\t" << children_connection[i] << "\t" << children_number[i]<< endl;
	}



	//cout << "norm = " << norm << endl;


	double r = u();
	double cumul = 0;
	int q = 0;
	bool children_flag = false;


	for(q=0; q< N_balls; q++){


		if(children_number[q]>0){
			if(q==c){
				double to_add = (children_number[q])*g;
				/*if(to_add<1 and children_number[q]>0) 
					to_add = 1.0;*/
				to_add = to_add/norm;			
				//if(to_add<1 and children_number[q]>0) to_add = 1;
				cumul+=to_add;					
			}
			else if(q==labels[c]){
				/*double to_add = children_number[q]*g;
				if(to_add<1 and children_number[q]>0) 
					to_add = 1.0;*/
				double to_add = (children_number[q])*g;
				to_add = to_add/norm;
				cumul+=to_add;						
			}
			else cumul+=children_number[q]*g*p/norm;
		}



		if(r< cumul){children_flag=true; break;}
		double w = weights[q];
		if(linked[q]) cumul+=w/norm;
		else if(labels[q]==labels[c]) cumul+=w/norm;
		else cumul+=w*gamma*f/norm;

		if(r< cumul) break;
	}



	//cout << q << "\t" <<r<< "\t" << cumul << "\t" << children_flag << "\t" << N_balls << endl;


	int extract_b;

	if(children_flag){
		children_number[q]--;
		if(children_number[q] < 0){
			cout << "Porco il clero" << endl;
			exit(-1);
		}


		add_ball(q);
		extract_b = N_balls - 1;
		reinforce_ball(extract_b, rho);
		exited[extract_b] = true;
		children_number[extract_b] =  nu + 1;
		last_visit.push_back(extract_b);
		creation_time.push_back(t);

	}
	else{
		extract_b = q;
		reinforce_ball(extract_b, rho);
	}	

	labels_count[labels[extract_b]+1]+=rho;

	/*cout << N_balls << "\t" << "Exctracted " << extract_b << endl;
	cout << "weight " << weights[extract_b]<< endl;
	cout << "children " << children_number[extract_b] << endl;*/
	//exit(-1);

	//if(t>0) exit(-1);
	return extract_b;
}

int Urn::print_adj_number(int c) {return adj_list[c].size();}

int Urn::get_balls_number(){
	int N_balls_tot = N_balls;
	//for(int i = 0; i < N_balls; i++) N_balls_tot+=children_number[i];
	return N_balls_tot;}

int Urn::get_exited_balls(){
	int N_exit = 0;
	for(int i = 0; i < N_balls; i++)
		if(exited[i]) N_exit++;

	return N_exit;
}

double Urn::get_max_weight(){
	double max_w = 0;
	for(int i = 0; i < N_balls; i++){
		if(max_w < weights[i]) max_w = weights[i];
	}
	return max_w;
}

void Urn::reset(int N0){

	N_balls = N0;
	weights.clear();
	adj_list.clear();
	exited.clear();
	balls.clear();

	for(int i = 0; i < N_balls; i++){
		balls.push_back(i);
		weights.push_back(1);
		adj_list.push_back(vector<int>(0));
		set_random_links_full(i);
		exited.push_back(false);
	}
}

vector<int> Urn::get_labels(){

	vector<int> counts(N_balls,0);

	for(int i = 0; i < N_balls; i++){
		int lab = labels[i];
		if(lab==-1) continue;
		counts[lab]+=1;
		if(counts[lab]>nu+1){
			cout << "Errore label " << lab << " figli rimasti " << children_number[lab] << endl;
			exit(-1);
		}
	}


	return labels;
}
